# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import hashlib
import os

from datetime import datetime
from functools import singledispatch
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Tuple, Union

from deker.types import Paths
from deker.types.private.enums import LocksExtensions


if TYPE_CHECKING:
    from deker.ABC.base_schemas import BaseAttributeSchema
    from deker.arrays import Array, VArray


def get_symlink_path(
    path_to_symlink_dir: Path,
    primary_attributes_schema: Optional[Tuple["BaseAttributeSchema", ...]],
    primary_attributes: dict,
) -> Path:
    """Make symlink path from given attributes.

    :param path_to_symlink_dir: path to array or varray symlinks in certain collection
    :param primary_attributes_schema: schemas of primary attributes
    :param primary_attributes: all primary attributes for a path
    """
    symlink_path = path_to_symlink_dir
    if primary_attributes_schema:
        for attr in primary_attributes_schema:
            attribute = primary_attributes[attr.name]
            if attr.name == "v_position":
                value = "-".join(str(el) for el in attribute)
            else:
                value = attribute.isoformat() if isinstance(attribute, datetime) else str(attribute)
            symlink_path = symlink_path / value
    return symlink_path


def get_main_path(array_id: str, data_directory: Path) -> Path:
    """Generate main path for the given array id by its type.

    :param array_id: Array or VArray id
    :param data_directory: Path to a certain data directory
    """
    main_tree, rest = array_id.split("-", 1)
    main_tree = os.path.sep.join(s for s in main_tree)
    return data_directory / main_tree / rest


def get_paths(array: Union["Array", "VArray"], collection_path: Union[Path, str]) -> Paths:
    """Create main and symlink paths for arrays.

    :param array: Array or VArray instance
    :param collection_path: path to a certain collection defined by user
    :returns: namedtuple with "main" and "symlink" pathlib.Paths
    """
    from deker.arrays import Array, VArray

    @singledispatch
    def get_path(arr: Union["Array", "VArray"], base_path: Path) -> Paths:  # noqa[ARG001]
        """Generate paths by object type.

        :param arr: any object
        :param base_path: collection base path
        """
        raise TypeError(f"Invalid object type: {type(arr)}")

    @get_path.register(Array)
    def array_path(arr: Array, base_path: Path) -> Paths:
        """Generate paths for Array.

        :param arr: Array type
        :param base_path: collection base path
        """
        # main_path
        main_path = get_main_path(arr.id, base_path / arr._adapter.data_dir)  # type: ignore[operator, attr-defined]

        # symlink
        primary_attributes: Optional[
            Tuple["BaseAttributeSchema", ...]
        ] = arr.schema.primary_attributes
        symlink_path = get_symlink_path(
            path_to_symlink_dir=base_path / arr._adapter.symlinks_dir,  # type: ignore[operator, attr-defined]
            primary_attributes_schema=primary_attributes,
            primary_attributes={
                **arr.primary_attributes,
                "vid": arr._vid,
                "v_position": arr._v_position,
            },
        )
        return Paths(main_path, symlink_path)

    @get_path.register(VArray)
    def varray_path(arr: VArray, base_path: Path) -> Paths:
        """Generate paths for VArray.

        :param arr: VArray type
        :param base_path: collection base path
        """
        # main_path
        main_path = get_main_path(arr.id, base_path / arr._adapter.data_dir)  # type: ignore[operator, attr-defined]

        # symlink
        primary_attributes: Optional[Tuple["BaseAttributeSchema"]] = arr.schema.primary_attributes
        symlink_path = get_symlink_path(
            path_to_symlink_dir=base_path / arr._adapter.symlinks_dir,  # type: ignore[operator, attr-defined]
            primary_attributes_schema=primary_attributes,
            primary_attributes={**arr.primary_attributes},
        )
        return Paths(main_path, symlink_path)

    coll_path = Path(collection_path) if isinstance(collection_path, str) else collection_path
    return get_path(array, coll_path)


def get_array_lock_path(array: Union["Array", "VArray"], data_directory_path: Path) -> Path:
    """Get array lockfile path.

    :param array: array to lock
    :param data_directory_path: path to data directory
    """
    file = (
        hashlib.md5(str(array.primary_attributes).encode()).hexdigest()
        + LocksExtensions.array_lock.value
    )
    return data_directory_path / file



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from datetime import datetime
from enum import Enum

import numpy as np


__all__ = ("DTypeEnum", "DimensionType", "LocksExtensions", "LocksTypes")


class DTypeEnum(Enum):
    """Mapping of numeric data type names to types."""

    int = int
    float = float
    complex = complex
    int8 = np.int8
    int16 = np.int16
    int32 = np.int32
    int64 = np.int64
    longlong = np.longlong
    float16 = np.float16
    float32 = np.float32
    float64 = np.float64
    float128 = np.float128
    longfloat = np.longfloat
    double = np.double
    longdouble = np.longdouble
    complex64 = np.complex64
    complex128 = np.complex128
    complex256 = np.complex256
    longcomplex = np.longcomplex
    string = str
    datetime = datetime
    tuple = tuple

    @staticmethod
    def get_name(object: "DTypeEnum") -> str:
        """Return object name.

        Numpy types parsing: we store types as "numpy.typename" in JSON.
        :param object: DtypeEnum object
        """
        if object.value in (int, float, complex, str, tuple, datetime):
            return str(object.name)
        return f"numpy.{object.name}"


class DimensionType(Enum):
    """Enum of dimensions' types."""

    generic = "generic"
    time = "time"


class LocksExtensions(Enum):
    """Extensions for lock files."""

    array_lock = ".arrlock"
    array_read_lock = ".arrayreadlock"
    collection_lock = ".lock"
    varray_lock = ".varraylock"


class LocksTypes(Enum):
    """Locks enum."""

    array_lock = "array creation lock"
    array_read_lock = "array read lock"
    collection_lock = "collection creation lock"
    varray_lock = "varray write lock"



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime

from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Type, Union

import numpy as np

from deker.dimensions import Dimension, TimeDimension
from deker.errors import DekerInvalidSchemaError
from deker.types import DTypeEnum, Numeric
from deker.validators import process_time_dimension_attrs


if TYPE_CHECKING:
    from deker.ABC.base_schemas import BaseDimensionSchema
    from deker.schemas import AttributeSchema, TimeDimensionSchema


def get_default_fill_value(dtype: Type[Numeric]) -> Any:
    """Get default fill value by dtype.

    :param dtype: schema dtype
    """
    if issubclass(dtype, np.inexact):
        return np.nan
    return np.iinfo(dtype).min  # type: ignore[type-var]


def create_dimensions(
    dimension_schemas: Tuple["BaseDimensionSchema", ...], attributes: Dict[str, Any]
) -> Tuple[Union[Dimension, TimeDimension], ...]:
    """Create Dimension and/or TimeDimension instances from dictionaries.

     Dimensions' and primary/custom attributes' schemas,
     stored as dictionaries in collection metadata, are converted into objects:
     Dimensions and/or TimeDimensions.

    :param dimension_schemas: a tuple of dimensions' schemas
    :param attributes: primary and custom attributes' joint dictionary
    """

    def create_time_dimension(tdim_schema: "TimeDimensionSchema") -> TimeDimension:
        """Create TimeDimension from schema.

        :param tdim_schema: TimeDimensionSchema
        """
        d = tdim_schema.as_dict
        if isinstance(d["start_value"], str):
            if d["start_value"].startswith("$"):
                d["start_value"] = process_time_dimension_attrs(attributes, d["start_value"][1:])
            else:
                d["start_value"] = datetime.datetime.fromisoformat(d["start_value"])

        if isinstance(d["step"], dict):
            d["step"] = datetime.timedelta(**d["step"])
        d.pop("type")
        return TimeDimension(**d)

    dimensions: List[Union[Dimension, TimeDimension]] = []
    for schema in dimension_schemas:
        if hasattr(schema, "start_value"):
            dimensions.append(create_time_dimension(schema))  # type: ignore[arg-type]
        else:
            dimensions.append(Dimension(**schema.as_dict))
    return tuple(dimensions)


def create_attributes_schema(attributes_schemas: List[dict]) -> Tuple["AttributeSchema", ...]:
    """Create AttributeSchema instances from a list of dictionaries.

    Primary and/or custom attributes' schemas, stored as dictionaries in collection metadata,
    are converted into AttributeSchema instances.

    :param attributes_schemas: a list of attributes' schemas dictionaries
    """
    from deker.schemas import AttributeSchema

    attributes = []
    try:
        for params in attributes_schemas:
            dtype = DTypeEnum[params["dtype"].lstrip("numpy.")].value
            attr_schema = AttributeSchema(**{**params, "dtype": dtype})
            attributes.append(attr_schema)
        return tuple(attributes)
    except (KeyError, ValueError, AttributeError) as e:
        raise DekerInvalidSchemaError(f'Schema "AttributeSchema" is invalid/corrupted : {e}')


def create_dimensions_schema(dimension_schemas: List[dict]) -> Tuple["BaseDimensionSchema", ...]:
    """Create DimensionSchema and/or TimeDimensionSchema instances from a list of dictionaries.

     Dimensions' schemas, stored as dictionaries in collection metadata, are converted into objects:
     DimensionSchema and/or TimeDimensionSchema.

    :param dimension_schemas: a list of dimensions' schemas dictionaries
    """
    from deker.schemas import DimensionSchema, TimeDimensionSchema

    schemas = []
    for dim in dimension_schemas:
        if "start_value" not in dim.keys():
            converted_dim = DimensionSchema(  # type: ignore[call-arg]
                name=dim["name"], size=dim["size"], labels=dim["labels"], scale=dim.get("scale")
            )
        else:
            try:
                start_value = datetime.datetime.fromisoformat(dim["start_value"])
            except ValueError:
                start_value = dim["start_value"]

            converted_dim = TimeDimensionSchema(  # type: ignore[call-arg]
                name=dim["name"],
                size=dim["size"],
                start_value=start_value,
                step=datetime.timedelta(**dim["step"]),
            )
        schemas.append(converted_dim)
    return tuple(schemas)



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


class DekerBaseApplicationError(Exception):
    """Base attribute exception."""


class DekerClientError(DekerBaseApplicationError, RuntimeError):
    """If something goes wrong in Client."""

    pass


class DekerArrayError(DekerBaseApplicationError):
    """If something goes wrong in Array."""


class DekerArrayTypeError(DekerArrayError):
    """If final Array's or VArray's values type do not match the dtype, set in Collection schema."""

    pass


class DekerCollectionError(DekerBaseApplicationError):
    """If something goes wrong in Collection."""

    pass


class DekerCollectionAlreadyExistsError(DekerCollectionError):
    """If collection with the same name already exists."""

    pass


class DekerCollectionNotExistsError(DekerCollectionError):
    """If collection doesn't exist."""

    pass


class DekerInvalidSchemaError(DekerBaseApplicationError):
    """If schema is invalid."""

    pass


class DekerMetaDataError(DekerInvalidSchemaError):
    """If metadata is invalid/corrupted."""

    pass


class DekerFilterError(DekerBaseApplicationError):
    """If something goes wrong during filtering."""

    pass


class DekerValidationError(DekerBaseApplicationError):
    """If something goes wrong during validation."""

    pass


class DekerIntegrityError(DekerBaseApplicationError):
    """If something goes wrong during integrity check."""

    pass


class DekerLockError(DekerBaseApplicationError):
    """If a Collection or a Array or VArray instance is locked."""

    pass


class DekerInstanceNotExistsError(DekerBaseApplicationError):
    """If instance was deleted, but user is trying to call methods on it."""

    pass


class DekerInvalidManagerCallError(DekerBaseApplicationError):
    """If we try to call method that shouldn't be called.

    E.g., .varrays in Arrays Collection
    """


class DekerSubsetError(DekerArrayError):
    """If something goes wrong while Subset managing."""


class DekerVSubsetError(DekerSubsetError):
    """If something goes wrong while VSubset managing."""


class DekerMemoryError(DekerBaseApplicationError, MemoryError):
    """Early memory overflow exception."""



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from functools import wraps
from typing import Any, Callable

from deker.errors import DekerClientError, DekerInstanceNotExistsError


def check_ctx_state(method: Callable) -> Callable:
    """Context checker deco.

    :param method: callable class method
    """

    @wraps(method)
    def ctx_checker(self: Any, *args: Any, **kwargs: Any) -> Any:
        """Inner executor methods wrapper.

        :param self: Adapter instance
        :param args: any
        :param kwargs: any
        """
        if self.ctx.is_closed:
            raise DekerClientError("Client is closed")
        return method(self, *args, **kwargs)

    return ctx_checker


def not_deleted(method: Callable) -> Callable:
    """Set rules of invocation for objects that could be deleted.

    If the object was deleted, do not call func.
    :param method: method to decorate
    """

    @wraps(method)
    def _inner(self: Any, *args: Any, **kwargs: Any) -> Any:
        """Check instance for flag.

        :param self: class instance
        :param args: any
        :param kwargs: any
        """
        if self._is_deleted():
            raise DekerInstanceNotExistsError(
                f"{self} doesn't exist, create new or get an instance again to be able to call {method}"
            )
        return method(self, *args, **kwargs)

    return _inner



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Abstract interfaces for subset and virtual subset."""

from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import TYPE_CHECKING, List, Optional, Tuple, Type, TypeVar, Union

import numpy as np

from deker_tools.slices import slice_converter


try:
    from xarray import DataArray

    xarray_import_error = None
except ImportError:
    DataArray = TypeVar("DataArray")
    xarray_import_error = ImportError(
        "No module named 'xarray' found (more likely due to it is not installed): "
        "try running `pip install xarray' or 'pip install deker[xarray]'"
    )

from deker.dimensions import Dimension, TimeDimension
from deker.errors import DekerSubsetError
from deker.log import SelfLoggerMixin
from deker.tools import check_memory
from deker.types.private.typings import Data, Numeric


if TYPE_CHECKING:
    from deker.ABC.base_adapters import BaseArrayAdapter, BaseVArrayAdapter
    from deker.ABC.base_array import BaseArray
    from deker.collection import Collection


class BaseSubset(SelfLoggerMixin, ABC):
    """A subset of Array or VArray data with bounds.

    It is the final object that can read, update and clear the data.
    """

    __slots__ = (
        "__bounds",
        "__shape",
        "__array",
        "__adapter",
        "__array_adapter",
        "__collection",
    )

    def __init__(
        self,
        slice_expression: Tuple[Union[slice, int], ...],
        shape: tuple,
        array: "BaseArray",
        array_adapter: "BaseArrayAdapter",
        varray_adapter: Optional["BaseVArrayAdapter"] = None,
        collection: Optional["Collection"] = None,
    ):
        """BaseSubset constructor.

        Sets adapters according to subset type: __array_adapter is used in VSubset only.

        :param slice_expression: a slice, tuple of slices or numpy IndexExpression, created by VArray.__getitem__()
        :param shape: subset shape, calculated by VArray.__getitem__()
        :param array: an instance of VArray
        :param array_adapter: ArrayAdapter instance
        :param varray_adapter: VArrayAdapter instances
        :param collection: Collection instance
        """
        self.__bounds = slice_expression
        self.__shape = shape
        self.__array = array
        self.__adapter = varray_adapter if varray_adapter else array_adapter
        self.__array_adapter = array_adapter if varray_adapter else None
        self.__collection = collection

    @property
    def shape(self) -> Tuple[int, ...]:
        """Subset shape."""
        return self.__shape

    @property
    def bounds(self) -> Tuple[Union[slice, int], ...]:
        """Subset bounds."""
        return self.__bounds

    @property
    def dtype(self) -> Type[Numeric]:
        """Subset dtype."""
        return self.__array.dtype

    @property
    def fill_value(self) -> Union[Numeric, type(np.nan)]:  # type: ignore[valid-type]
        """Get subset fill_value."""
        return self.__array.fill_value

    @abstractmethod
    def read(self) -> Union[Numeric, np.ndarray, None]:
        """Read data from array slice."""
        pass

    def read_xarray(self) -> DataArray:
        """Read data from Array as xarray.DataArray.

        DataArray is an object of `xarray` (https://docs.xarray.dev) library,
        which contains subset data and its description. It provides conversion
        to different formats including pandas.Dataframe, netCDF, Zarr and many others.

        However, `xarray` library is not a Deker's dependency. If you wish to use
        this method you shall install xarray manually or run `pip install deker[xarray]`.

        Subset's description is made before the data is read.
        Right after the description is created - memory checker is invoked.
        If your RAM is insufficient for reading data - `DekerMemoryError` is raised.

        Warning: final xarray.DataArray data shape may differ from subset shape
        """
        if xarray_import_error:
            raise xarray_import_error
        subset_name = self.__class__.__name__
        if self.shape == ():
            raise DekerSubsetError(
                f"Cannot convert a scalar deker.{subset_name} to xarray.DataArray"
            )
        description = self.describe()
        dims = list(description)
        new_shape = tuple(len(description[dim]) for dim in description)
        check_memory(self.shape, self.dtype, self.__adapter.ctx.config.memory_limit)
        data = self.read()
        data = data.reshape(new_shape)

        data_array = DataArray(
            name=f"Deker {subset_name} of {self.__array.__class__.__name__}(id={self.__array.id}) "
            f"from collection `{self.__array.collection}`",
            data=data,
            dims=dims,
            coords=description,
            attrs={
                "primary_attributes": self.__array.primary_attributes,
                "custom_attributes": self.__array.custom_attributes,
            },
        )
        return data_array

    @abstractmethod
    def update(self, data: Data) -> None:
        """Update data in array by slice.

        :param data: new data which shall match subset slicing
        """
        pass

    @abstractmethod
    def clear(self) -> None:
        """Clear data in array by slice."""
        pass

    def _is_deleted(self) -> bool:
        """Check if array was deleted."""
        return self.__adapter.is_deleted(self.__array)

    @staticmethod
    def __generate_full_scale(dimension: Dimension) -> List:
        """Generate full scale of scaled dimension.

        Floats, autogenerated with np.arange, may differ both from set step
        and from expected final quantity:

         - np.arange(1.5, 1.6, 0.1)
            array([1.5, 1.6]) (expected just [1.5])

         - np.arange(-1.0, 0.5, 0.1)
            array([-1.00000000e+00, -9.00000000e-01, -8.00000000e-01, -7.00000000e-01,
            -6.00000000e-01, -5.00000000e-01, -4.00000000e-01, -3.00000000e-01,
            -2.00000000e-01, -1.00000000e-01, -2.22044605e-16,  1.00000000e-01,
             2.00000000e-01,  3.00000000e-01,  4.00000000e-01])

             (expected just array([-1.0, -0.9, -0.8, -0.7, -0.6,
                            -0.5, -0.4, -0.3, -0.2, -0.1,
                            0.0,  0.1, 0.2,  0.3,  0.4]))

        Regarding that all values are floats, we shall generate them carefully
        and try to round their fractional part to digits quantity similar
        to start_value attribute's fractional part.

        :param dimension: dimension with non-empty `scale` attribute
        """
        step = dimension.scale.step
        start_value = dimension.scale.start_value
        end = start_value + dimension.size * step
        rounder = len(str(step).split(".")[-1])
        value = np.around(np.arange(start_value, end, dimension.scale.step), rounder)
        return value.tolist()[: dimension.size]  # type: ignore[return-value]

    def __describe_full_dimension(self, dimension: Union[TimeDimension, Dimension]) -> list:
        """Generate dimension's full range of descriptive values.

        :param dimension: dimension of any Deker type
        """
        if isinstance(dimension, TimeDimension):
            value = [
                dimension.start_value + n * dimension.step for n in range(dimension.size)  # type: ignore[operator]
            ]

        elif dimension.labels:
            value = list(dimension.labels)
        elif dimension.scale:
            value = self.__generate_full_scale(dimension)
        else:  # just indexes
            value = list(range(0, dimension.size, dimension.step))
        return value

    def __describe_slice(self, dimension: Union[TimeDimension, Dimension], bound: slice) -> list:
        """Generate dimension's range of descriptive values by slice bounds.

        :param dimension: dimension of any Deker type
        :param bound: slice object
        """
        if bound == slice(None, None, None):
            return self.__describe_full_dimension(dimension)

        # convert slice start value to positive integer
        if bound.start is None:
            start = 0
        elif bound.start >= 0:
            start = bound.start
        else:
            start = dimension.size + bound.start

        # convert slice stop value to positive integer
        if bound.stop is None:
            stop = dimension.size
        elif bound.stop >= 0:
            stop = bound.stop
        else:
            stop = dimension.size + bound.stop

        # ignore slice step value
        step = dimension.step

        if isinstance(dimension, TimeDimension):
            start_value = dimension.start_value + step * start  # type: ignore[operator]
            end = stop - start
            value = [start_value + n * step for n in range(end)]  # type: ignore[operator]

        elif dimension.labels:
            value = list(dimension.labels[bound])
        elif dimension.scale:
            value = self.__generate_full_scale(dimension)[bound]
        else:  # just indexes
            start_value = start if start is not None else 0
            end = stop
            value = list(range(start_value, end, step))  # type: ignore[arg-type]
        return value

    def describe(self) -> OrderedDict:
        """Describe subset data.

        You shall remember that once it is described, your physical memory becomes reduced
        by the size of all descriptive values of all the subset's dimensions.
        In other words, it equals (more or less) to the size of your data that you are going to manage.
        And if you keep this description in memory, your RAM may become insufficient for data managing.

        So, we highly recommend using this method in debug or deleting its result manually right before
        invocation of any data-managing methods.

        But still subset's description is automatically created and inserted to xarray.DataArray
        on `read_xarray()` call.

        If your RAM is insufficient for reading data - DekerMemoryError is raised.
        """
        description = OrderedDict()
        for n, bound in enumerate(self.bounds):
            dimension = self.__array.dimensions[n]
            if isinstance(bound, slice):
                value = self.__describe_slice(dimension, bound)
            else:
                # convert bound value to positive integer
                bound = bound if bound >= 0 else dimension.size + bound

                if isinstance(dimension, TimeDimension):
                    value = [dimension.start_value + dimension.step * bound]  # type: ignore[operator]

                elif dimension.labels:
                    value = [dimension.labels[bound]]
                elif dimension.scale:
                    value = [dimension.scale.start_value + bound * dimension.scale.step]
                else:  # just index
                    value = [bound]
            description.update({dimension.name: value})

        # add full description for dimensions if they were excluded from bounds
        dims = self.__array.dimensions
        if len(self.bounds) < len(dims):  # type: ignore[arg-type]
            start_idx = len(dims) - (len(dims) - len(self.bounds))  # type: ignore[arg-type]
            for dimension in dims[start_idx:]:
                description.update({dimension.name: self.__describe_full_dimension(dimension)})

        return description

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"bounds={slice_converter[self.__bounds]}, shape={self.shape}, array={self.__array.id}"
            f")"
        )

    def __str__(self) -> str:
        return f"{self.__array!s}{slice_converter[self.__bounds]}"



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fcntl

from pathlib import Path
from typing import Any

from deker.errors import DekerLockError
from deker.log import SelfLoggerMixin
from deker.types.private.enums import LocksExtensions


class Flock(SelfLoggerMixin):
    """File locker.

    :param file: path to the file to be locked
    """

    def __init__(self, file: Path) -> None:
        self.file = file
        self.fd = None  # file descriptor, closed on Flock release
        self.logger.debug("Instantiated")

    def acquire(self) -> None:
        """Create and lock lockfile."""
        # If the file doesn't end with .lock or .arrlock,
        # then it's array or collection lock (e.g .json/.hdf)
        self.logger.debug(f"Trying to acquire lock for {self.file}")
        if str(self.file).endswith(
            (LocksExtensions.array_lock.value, LocksExtensions.collection_lock.value)
        ):
            if not self.file.exists():
                self.file.parent.mkdir(parents=True, exist_ok=True)
            mode = "w"
        else:
            mode = "r"
        try:
            self.fd = open(self.file, mode=mode)  # noqa[PTH123, SIM115]
            self.logger.debug(f"Opened descriptor for {self.file}")
            fcntl.flock(self.fd, fcntl.LOCK_EX | fcntl.LOCK_NB)  # type: ignore[arg-type]
            self.logger.debug(f"{self.file} acquired lock")
        except FileNotFoundError:
            self.logger.debug(f"{self.file} not found")
        except BlockingIOError:
            if self.fd:
                self.fd.close()
                self.logger.debug(f"Closed descriptor for {self.file} due to BlockingIOError")
            raise DekerLockError(f"{self.file} is locked")

    def release(self) -> None:
        """Releases lockfile."""
        self.logger.debug(f"trying to release lock for {self.file}")
        if self.fd:
            fcntl.flock(self.fd, fcntl.LOCK_UN)
            self.fd.close()
            self.logger.debug(f"{self.file} released lock")
        if self.file.name.endswith(LocksExtensions.array_lock.value):
            self.file.unlink(missing_ok=True)
            self.logger.debug(f"{self.file} unlinked symlink")

    def __enter__(self) -> "Flock":
        self.acquire()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        return self.release()



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from deker.client import Client
from deker.log import set_logging_level
from deker.schemas import (
    ArraySchema,
    AttributeSchema,
    DimensionSchema,
    TimeDimensionSchema,
    VArraySchema,
)
from deker.types import Scale


__all__ = (  # F405
    "Client",
    "DimensionSchema",
    "TimeDimensionSchema",
    "ArraySchema",
    "AttributeSchema",
    "VArraySchema",
    "Scale",
    "set_logging_level",
)

# shell_completions are used in deker-shell to exclude Deker objects
# from autocompletion of parameters during initialization
shell_completions = tuple(deker_obj + "(" for deker_obj in __all__)



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

import tqdm

from deker.ABC.base_integrity import BaseChecker
from deker.arrays import Array, VArray
from deker.collection import Collection
from deker.errors import (
    DekerBaseApplicationError,
    DekerCollectionNotExistsError,
    DekerIntegrityError,
)
from deker.tools import get_main_path, get_symlink_path
from deker.types.private.enums import LocksExtensions


if TYPE_CHECKING:
    from deker.client import Client


class DataChecker(BaseChecker):
    """Checks Array's single number data from subset."""

    CHECKER_LEVEL = 4

    def check(self, array: Array) -> None:
        """Check if array can be read correctly.

        :param array: array to check
        """
        if self.level < self.CHECKER_LEVEL:
            return

        try:
            subset = array[tuple(s - 1 for s in array.shape)]
            data = subset.read()
        except Exception as e:
            raise DekerIntegrityError(f"Array {array.id} data is corrupted: {e!s}")
        if data.dtype != array.dtype:
            raise DekerIntegrityError(f"Array {array.id} data is corrupted: incorrect dtype")


class PathsChecker(BaseChecker):
    """Checks Array or VArray paths."""

    CHECKER_LEVEL = 3

    def _validate_path_array_without_key_attributes(
        self, main_path: Path, symlink_path: Path, files: list
    ) -> None:
        """Validate symlinks in array without key attributes.

        :param main_path: path to the array data
        :param symlink_path: path to symlink directory
        :param files: files in symlink directory
        """
        for file in files:
            if main_path / file == Path.readlink(symlink_path / file):
                self.paths[main_path] = symlink_path
                break
        if not self.paths.get(main_path):
            raise DekerIntegrityError(f"Incorrect symlink {symlink_path}")

    def _validate_paths(self, main_path: Path, symlink_path: Path, collection: Collection) -> None:
        """Validate symlink and main paths.

        :param main_path: Array or VArray main path
        :param symlink_path: Array or VArray symlink
        :param collection: Collection
        """
        if not symlink_path.exists():
            raise DekerIntegrityError(f"Symlink {symlink_path} not found")

        files = [str(f.name) for f in symlink_path.iterdir()]

        if len(files) < 1:
            raise DekerIntegrityError(f"Symlink {symlink_path} not found")

        if collection.array_schema.primary_attributes:
            if len(files) > 1:
                files = ["\t- " + str(file) + "\n" for file in files]
                raise DekerIntegrityError(
                    f"There are unnecessary files in directory:\n{symlink_path}\n {''.join(files)}"
                )

            if main_path / files[0] == Path.readlink(symlink_path / files[0]):
                self.paths[main_path] = symlink_path
            else:
                raise DekerIntegrityError(f"Incorrect symlink {symlink_path}")
        else:
            self._validate_path_array_without_key_attributes(main_path, symlink_path, files)

    def check(self, array: Union[Array, VArray], collection: Collection) -> None:
        """Check symlink and main paths.

        :param array: array to check
        :param collection: Collection
        """
        if self.level < self.CHECKER_LEVEL:
            return
        if isinstance(array, Array):
            main_path_args = (array.id, collection.path / self.ctx.config.array_data_directory)
            symlink_dir = self.ctx.config.array_symlinks_directory
            symlink_path_args = (
                collection.path / symlink_dir,
                array.schema.primary_attributes,
                {
                    **array.primary_attributes,
                    "vid": array._vid,
                    "v_position": array._v_position,
                },
            )
        else:
            main_path_args = (array.id, collection.path / self.ctx.config.varray_data_directory)
            symlink_dir = self.ctx.config.varray_symlinks_directory
            symlink_path_args = (
                collection.path / symlink_dir,
                array.schema.primary_attributes,
                {**array.primary_attributes},
            )
        main_path = get_main_path(*main_path_args)
        symlink_path = get_symlink_path(*symlink_path_args)
        self._validate_paths(main_path, symlink_path, collection)

        if self.next_checker:
            self.next_checker.check(array)


class ArraysChecker(BaseChecker):
    """Checks arrays in collection."""

    CHECKER_LEVEL = 2

    def check_arrays_locks(self, collection: Collection) -> None:
        """Check if Arrays or VArrays have no lockfiles left from create method.

        :param collection: Collection to be checked
        """
        for file in Path.rglob(collection.path, "*lock"):
            if file.name.endswith(LocksExtensions.array_lock.value):
                self.errors[f"Collection {collection.name} array create locks:"].append(file.name)
            elif file.name.endswith(LocksExtensions.array_read_lock.value):
                self.errors[
                    f"Collection {collection.name} array read locks are detected. Use "
                    f"client.clear_locks:"
                ].append(file.name)
            elif file.name.endswith(LocksExtensions.varray_lock.value):
                self.errors[
                    f"Collection {collection.name} varray write locks are detected. Use "
                    f"client.clear_locks:"
                ].append(file.name)
        if self.stop_on_error and self.errors:
            raise DekerIntegrityError(self._parse_errors())

    def check(self, collection: Collection) -> None:
        """Check if Arrays or VArray in Collection are valid.

        :param collection: Collection to be checked
        """
        if self.level < self.CHECKER_LEVEL:
            return
        self.check_arrays_locks(collection)

        for array in collection.arrays:
            try:
                if self.next_checker:
                    self.next_checker.check(array, collection)
            except DekerBaseApplicationError as e:
                if not self.stop_on_error:
                    self.errors[f"Collection {collection.name} arrays integrity errors:"].append(
                        str(e)
                    )
                else:
                    raise DekerIntegrityError(str(e))
        if collection.varray_schema:
            for varray in collection.varrays:
                try:
                    if self.next_checker:
                        self.next_checker.check(varray, collection)
                except DekerBaseApplicationError as e:
                    if not self.stop_on_error:
                        self.errors[
                            f"Collection {collection.name} varrays integrity errors:"
                        ].append(str(e))
                    else:
                        raise DekerIntegrityError(str(e))


class CollectionsChecker(BaseChecker):
    """Checks collections initialization & lockfiles."""

    CHECKER_LEVEL = 1

    def check_collections(self) -> list:
        """Check collections integrity."""
        collections: list = []
        locks: list = []
        for directory in Path(self.root_path).iterdir():
            try:
                if directory.is_file() and directory.name.endswith(
                    LocksExtensions.collection_lock.value
                ):
                    locks.append(directory.name[: -len(LocksExtensions.collection_lock.value)])
                collection = self.client.get_collection(directory.name)
                if collection:
                    collections.append(collection)
            except DekerBaseApplicationError as e:
                self.errors["Collections initialization errors:"].append(str(e))
        for collection in collections:
            if collection.name not in locks:
                self.errors["Collections locks errors:"].append(
                    f"BaseLock for {collection.name} not found"
                )
        for lock in locks:
            for collection in collections:
                if collection.name == lock:
                    break
            else:
                self.errors["Collections locks errors:"].append(
                    f"Collection with lock {lock} not found"
                )
        if self.stop_on_error and self.errors:
            raise DekerIntegrityError(self._parse_errors())
        return collections

    def check(self, collection_name: Optional[str] = None) -> None:
        """Check collections and run integrity check for every collection if level > 1.

        :param collection_name: optional collection to be checked
        """
        if collection_name:
            # skipping collections checker
            self.next_checker: Optional[BaseChecker] = (
                self.next_checker.next_checker if self.next_checker else None
            )
            try:
                collection: Collection = self.client.get_collection(collection_name)
                if not collection:
                    raise DekerCollectionNotExistsError(
                        f"Collection {collection_name} does not exist at this storage"
                    )
                if self.next_checker:
                    self.next_checker.check(collection)
            except DekerCollectionNotExistsError:
                raise
            except Exception as e:
                self.errors["Collections initialization errors:"].append(str(e))
                return
        collections = self.check_collections()
        if self.level > self.CHECKER_LEVEL:
            collections_pbar = tqdm.tqdm(collections)
            for collection in collections_pbar:
                collections_pbar.set_description(f'Checking collection "{collection.name}"')
                if self.next_checker:
                    self.next_checker.check(collection)


class IntegrityChecker(BaseChecker):
    """Storage integrity checker."""

    # :
    def __init__(self, client: "Client", root_path: Path, stop_on_error: bool, level: int) -> None:
        super().__init__(stop_on_error, {}, defaultdict(list), level, client, root_path)  # type: ignore[attr-defined]
        self.add_checker(CollectionsChecker)
        self.add_checker(ArraysChecker)
        self.add_checker(PathsChecker)
        self.add_checker(DataChecker)

    def check(self, collection_name: Optional[str] = None) -> str:
        """Run integrity check.

        :param collection_name: optional collection to check
        """
        print("Integrity check is running...")
        if self.next_checker:
            self.next_checker.check(collection_name)
        return self._parse_errors()



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module for using in type annotations. Not for direct initialization."""
from datetime import datetime
from typing import TYPE_CHECKING, List, Tuple, Type, Union

import numpy as np

from numpy.lib.index_tricks import IndexExpression

from .classes import ArrayPosition, ArraysCoordinatesWithOffset


__all__ = (
    "EllipsisType",
    "NoneType",
    "Arrays",
    "ArraysCoordinates",
    "ArraysInDimension",
    "Slice",
    "FancySlice",
    "MetaDataType",
    "Data",
    "Numeric",
    "NumericDtypes",
)

EllipsisType = type(...)
NoneType = type(None)

if TYPE_CHECKING:
    EllipsisType = Type[Ellipsis]

NumericDtypes = [
    int,
    float,
    complex,
    np.int_,  # alias for np.compat.long, deprecated in numpy version 1.25, equals to np.int64
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.float16,
    np.float64,
    np.float128,
    np.longfloat,
    np.double,
    np.longdouble,
    np.complex64,
    np.complex128,
    np.complex256,
    np.longcomplex,
    np.longlong,
]

Numeric = Union[
    int,
    float,
    complex,
    np.int_,  # alias for np.compat.long, deprecated in numpy version 1.25, equals to np.int64
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.float16,
    np.float64,
    np.float128,
    np.longfloat,
    np.double,
    np.longdouble,
    np.complex64,
    np.complex128,
    np.complex256,
    np.longcomplex,
    np.longlong,
]

Data = Union[list, tuple, np.ndarray, Numeric]
MetaDataType = Union[int, float, str, datetime, Data]
Slice = Union[
    IndexExpression,
    slice,
    EllipsisType,  # type: ignore
    int,
    Tuple[Union[slice, int, EllipsisType, None], ...],  # type: ignore
]

FancySlice = Union[
    IndexExpression,
    slice,
    EllipsisType,  # type: ignore
    int,
    float,
    str,
    datetime,
    None,
    Tuple[Union[IndexExpression, slice, int, float, str, datetime, EllipsisType, None], ...],  # type: ignore
]

ArraysInDimension = List[ArraysCoordinatesWithOffset]
ArraysCoordinates = List[ArraysInDimension]
Arrays = List[ArrayPosition]



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Abstract interfaces for collection."""

from abc import ABC, abstractmethod
from typing import Any, Optional

from deker.types.private.classes import Serializer


class BaseCollectionOptions(Serializer, ABC):
    """Base interface for collection options.

    Options, such as chunks, compression -
    or whatever that may somehow influence storage files' size or structure -
    depend on provided storage adapter
    """

    chunks: Any
    compression_opts: Any

    @classmethod
    @abstractmethod
    def _process_options(cls, storage_options: Optional[dict]) -> dict:
        """Validate and convert collection storage options.

        :param storage_options: options for storing data by a certain storage adapter,
        like chunks, compression, etc. Such options are passed to a collection to be used for every
        Array in it.
        """
        pass



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Abstract interfaces for adapters."""

import datetime
import logging

from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, List, Optional, Tuple, Type, Union, get_args

import numpy as np

from deker_tools.slices import create_shape_from_slice
from deker_tools.time import get_utc
from numpy import ndarray

from deker.ctx import CTX
from deker.errors import (
    DekerArrayTypeError,
    DekerFilterError,
    DekerMetaDataError,
    DekerValidationError,
)
from deker.log import SelfLoggerMixin
from deker.schemas import SchemaTypeEnum
from deker.tools import create_attributes_schema, create_dimensions_schema
from deker.tools.decorators import check_ctx_state
from deker.types import DTypeEnum
from deker.types.private.classes import ArrayMeta
from deker.types.private.typings import Data, EllipsisType, Numeric, Slice


if TYPE_CHECKING:
    from deker.ABC.base_array import BaseArray
    from deker.ABC.base_collection import BaseCollectionOptions
    from deker.ABC.base_factory import BaseAdaptersFactory
    from deker.ABC.base_schemas import BaseArraysSchema
    from deker.arrays import Array, VArray
    from deker.collection import Collection
    from deker.uri import Uri


class BaseStorageAdapter(ABC):
    """Interface for adapters working with files with data (e.g hdf5, tiff, etc).

    Doesn't accept any params for instantiation.
    """

    file_ext: str
    storage_options: Type["BaseCollectionOptions"]

    @classmethod
    @abstractmethod
    def create(
        cls, path: Path, array_shape: Tuple[int, ...], metadata: Union[str, bytes, dict]
    ) -> None:
        """Create a new hdf5 file with metadata.

        :param path: path to hdf5 file
        :param array_shape: shape of the array
        :param metadata: array metadata
        """
        pass

    @classmethod
    @abstractmethod
    def read_data(
        cls,
        path: Path,
        array_shape: Tuple[int, EllipsisType],  # type: ignore
        bounds: Slice,
        fill_value: Numeric,
        dtype: Type[Numeric],
    ) -> ndarray:
        """Read array data from hdf5 file.

        :param path: path to hdf5 file
        :param array_shape: shape of the array
        :param bounds: array slice
        :param fill_value: value to fill the empty array
        :param dtype: array dtype
        """
        pass

    @classmethod
    @abstractmethod
    def read_meta(cls, path: Path) -> ArrayMeta:
        """Read array metadata from hdf5 file.

        :param path: path to hdf5 file
        """
        pass

    @classmethod
    @abstractmethod
    def update_data(
        cls,
        path: Path,
        bounds: Slice,
        data: Any,
        dtype: Type[Numeric],
        shape: tuple,
        fill_value: Numeric,
        collection_options: Optional["BaseCollectionOptions"],
    ) -> None:
        """Update array data in hdf5 file.

        :param path: path to hdf5 file
        :param bounds: array slice
        :param data: new data for array slice
        :param dtype: array dtype
        :param shape: array shape
        :param fill_value: array fill_value
        :param collection_options: chunking and compression options
        """
        pass

    @classmethod
    @abstractmethod
    def update_meta_custom_attributes(cls, path: Path, attributes: dict) -> dict:
        """Update metadata in the hdf5 file.

        :param path: path to hdf5 file
        :param attributes: new custom attributes
        """
        pass

    @classmethod
    @abstractmethod
    def clear_data(
        cls, path: Path, array_shape: Tuple[int, ...], bounds: Slice, fill_value: Numeric
    ) -> bool:
        """Clear array data in hdf5 file.

        :param path: path to hdf5 file
        :param array_shape: array shape
        :param bounds: array bounds to update
        :param fill_value: array fill_value
        """
        pass


class BaseCollectionAdapter(SelfLoggerMixin, ABC):
    """Base interface for collection adapters."""

    file_ext: str
    metadata_version: str

    def __init__(self, ctx: CTX):
        self.ctx = ctx
        self.uri = self.ctx.uri

    def __del__(self) -> None:
        """Delete object."""
        self.close()

    @property
    @abstractmethod
    def collections_resource(self) -> Path:
        """Return a path to collections' resource."""
        pass

    def close(self) -> None:
        """Close adapter."""
        self.uri = ""

    @abstractmethod
    def create(self, collection: "Collection") -> None:
        """Add collection to storage and creates file structure.

        :param collection: Collection to be created
        """
        pass

    @abstractmethod
    def read(self, name: str) -> dict:
        """Read collection metadata.

        :param name: Collection name
        """
        pass

    @abstractmethod
    def delete(self, collection: "Collection") -> None:
        """Delete collection.

        :param collection: Collection to delete
        """
        pass

    @abstractmethod
    def clear(self, collection: "Collection") -> None:
        """Clear collection directory.

        :param collection: Collection to be cleared
        """
        pass

    @check_ctx_state
    def create_collection_from_meta(
        self, collection_data: dict, factory: "BaseAdaptersFactory"
    ) -> "Collection":
        """Create a Collection instance from meta information.

        :param collection_data: dict with information about Collection (schemas, vgrid, etc)
        :param factory: adapters factory instance
        """
        from deker.collection import Collection

        name: str = collection_data["name"]
        data: dict = collection_data["schema"]

        # Check schema type
        schema_class = SchemaTypeEnum[collection_data.get("type")].value

        try:
            dtype = DTypeEnum[data["dtype"].lstrip("numpy.")].value
            fill_value = (
                dtype(data["fill_value"]) if data["fill_value"] is not None else data["fill_value"]
            )
            # Create schema
            schema = schema_class(
                **{
                    **data,
                    "dimensions": create_dimensions_schema(data["dimensions"]),
                    "attributes": create_attributes_schema(data["attributes"]),
                    "dtype": dtype,
                    "fill_value": fill_value,
                }
            )
            storage_adapter = self.get_storage_adapter(collection_data.get("storage_adapter"))
            coll_params = storage_adapter.storage_options._process_options(
                collection_data.get("options")
            )

            collection_options = (
                storage_adapter.storage_options(**coll_params) if coll_params else None
            )

            collection = Collection(
                name=name,
                schema=schema,
                adapter=self,
                factory=factory,
                collection_options=collection_options,
                storage_adapter=storage_adapter,
            )

            logging.debug(f"Collection {collection.name} read")
            return collection
        except (KeyError, ValueError, TypeError, AttributeError) as e:
            logging.exception(e)  # noqa[TRY401]
            raise DekerMetaDataError(f'Collection "{name}" metadata is invalid/corrupted: {e}')

    @staticmethod
    def get_storage_adapter(storage_adapter: Optional[str] = None) -> Type[BaseStorageAdapter]:
        """Get storage adapter type.

        :param storage_adapter: Storage adapter name
        """
        from deker_local_adapters import storage_adapter_factory

        return storage_adapter_factory(storage_adapter)

    @abstractmethod
    def __iter__(self) -> "Collection":
        """Iterate over all Collections in the storage."""
        pass


class IArrayAdapter(ABC):
    """Interface for Array  or VArray adapters."""

    def __init__(
        self,
        collection_path: Union["Uri", Path],
        ctx: CTX,
        executor: ThreadPoolExecutor,
        storage_adapter: Type[BaseStorageAdapter],
    ) -> None:
        self.collection_path = collection_path
        self.ctx = ctx
        self.executor = executor
        self.uri = self.ctx.uri
        self.storage_adapter = storage_adapter()

    @staticmethod
    def _process_data(
        array_dtype: Type[Numeric],
        array_shape: tuple,
        data: Data,
        bounds: Optional[Slice] = None,
    ) -> Union[ndarray, Numeric]:
        """Validate data over array dtype and bounds.

        :param array_dtype: array dtype
        :param array_shape: array shape
        :param data: data to be validated
        :param bounds: array bounds
        """
        if not isinstance(data, (list, tuple, np.ndarray, get_args(Numeric))):  # type: ignore[arg-type]
            raise DekerArrayTypeError(f"Invalid data type: {type(data)}; {array_dtype} expected")

        if isinstance(data, get_args(Numeric)):  # type: ignore[arg-type]
            data_type = type(data)
            if data_type != array_dtype:
                cdata = array_dtype(data)  # type: ignore[arg-type]
                if data_type(cdata) != data:  # type: ignore[arg-type,call-overload]
                    raise DekerArrayTypeError(
                        f"Invalid data type: {type(data)}; {array_dtype} expected"
                    )
                data = cdata
        else:
            sliced_bounds = (
                np.index_exp[:]
                if bounds is None or bounds == slice(None, None, None)
                else np.index_exp[bounds]  # type: ignore[type-var]
            )
            shape = create_shape_from_slice(array_shape, sliced_bounds)  # type: ignore[arg-type]

            if isinstance(data, (list, tuple)):
                try:
                    data = np.asarray(data).astype(dtype=array_dtype, casting="no")
                except TypeError as e:
                    raise DekerArrayTypeError(f"Invalid data type: {e}; {array_dtype} expected")

            if isinstance(data, ndarray):
                if data.dtype != array_dtype:
                    raise DekerArrayTypeError(
                        f"Invalid data dtype: {data.dtype}; {array_dtype} expected"
                    )
                if shape != data.shape:
                    raise IndexError(f"Invalid data shape: {data.shape}; {shape} expected")
        return data  # type: ignore[return-value]

    @abstractmethod
    def create(self, array: Union[dict, "BaseArray"]) -> Union["Array", "VArray"]:
        """Create Array or VArray in the database.

        :param array: instance of Array or VArray
        """
        pass

    @abstractmethod
    def read_meta(self, array: Union["BaseArray", Path]) -> ArrayMeta:
        """Read array metadata.

        :param array: Array or VArray instance or Path to meta file
        """
        pass

    @abstractmethod
    def update_meta_custom_attributes(self, array: "BaseArray", attributes: dict) -> None:
        """Update custom attributes in meta.

        :param array: Array or VArray
        :param attributes: attributes to be updated
        """
        pass

    @abstractmethod
    def delete(
        self,
        array: "BaseArray",
    ) -> None:
        """Delete a whole array.

        :param array: Array or VArray
        """
        pass

    @abstractmethod
    def read_data(self, array: "BaseArray", bounds: Slice) -> Union[Numeric, ndarray]:
        """Read data from a section of array.

        :param array: Array or VArray
        :param bounds: bounds of a section to be read
        """
        pass

    @abstractmethod
    def update(self, array: "BaseArray", bounds: Slice, data: Any) -> None:
        """Update a section of array.

        :param array: Array or VArray
        :param bounds: bounds of a section to be updated
        :param data: new data to be inserted into Array
        """
        pass

    @abstractmethod
    def clear(self, array: "BaseArray", bounds: Slice) -> None:
        """Clear a section of array.

        :param array: Array or VArray
        :param bounds: bounds of a section to be cleared
        """
        pass

    @abstractmethod
    def is_deleted(self, array: "BaseArray") -> bool:
        """Check if array was deleted.

        :param array: Array to check
        """
        pass

    def filter(
        self,
        filters: dict,
        schema: "BaseArraysSchema",
        collection: "Collection",
        array_adapter: "BaseArrayAdapter",
        varray_adapter: Optional["BaseVArrayAdapter"],
    ) -> List[Union["Array", "VArray"]]:
        """Filter arrays.

        :param filters:  dict with filtering criteria
        :param schema: Array or Varray schema
        :param collection: Collection instance
        :param array_adapter: Arrays' adapter
        :param varray_adapter: VArrays' adapter
        """
        # If we got "id" in filters, just return the array with the given ID
        # (Till AND/OR logic not implemented)
        if "id" in filters:
            array = self.get_by_id(filters["id"], collection, array_adapter, varray_adapter)
            return [array] if array else []

        primary_attribute_filters = {}
        error_message = "Filtering by non primary attributes is not implemented yet"
        # If there are no primary attributes in the schema
        if not schema.primary_attributes:
            raise NotImplementedError(error_message)

        for primary_attr in schema.primary_attributes:
            value = filters.get(primary_attr.name)
            # If any of key attrs wasn't passed, use simple filter
            if value is None:
                raise NotImplementedError(error_message)
            if primary_attr.dtype == datetime.datetime:
                if not isinstance(value, (str, datetime.datetime)):
                    raise DekerValidationError(
                        f"Invalid type passed for {primary_attr.name} filtering: {type(value)}; "
                        f"only datetime.datetime or datetime iso-string are allowed"
                    )
                value = get_utc(value)
            primary_attribute_filters[primary_attr.name] = value

        # If extra args were passed
        if len(filters.keys()) > len(primary_attribute_filters.keys()):
            raise DekerFilterError("Some arguments don't exist in schema")

        # Filter by primary attributes

        array = self.get_by_primary_attributes(
            primary_attribute_filters, schema, collection, array_adapter, varray_adapter
        )
        if array:
            return [array]
        return []

    @abstractmethod
    def get_by_id(
        self,
        id_: str,
        collection: "Collection",
        array_adapter: "BaseArrayAdapter",
        varray_adapter: Optional["BaseVArrayAdapter"],
    ) -> Optional[Union["Array", "VArray"]]:
        """Find Array or VArray  by id.

        :param id_: Array id
        :param collection: Collection
        :param array_adapter: Arrays' adapter
        :param varray_adapter: VArrays' adapter
        """
        pass

    @abstractmethod
    def get_by_primary_attributes(
        self,
        primary_attributes: dict,
        schema: "BaseArraysSchema",
        collection: "Collection",
        array_adapter: "BaseArrayAdapter",
        varray_adapter: Optional["BaseVArrayAdapter"],
    ) -> Optional[Union["Array", "VArray"]]:
        """Find Array or VArray by given primary attributes.

        :param primary_attributes: key attributes
        :param schema: Array or VArray schema
        :param collection: Collection instance
        :param array_adapter: Arrays' adapter
        :param varray_adapter: VArrays' adapter
        """
        pass

    @abstractmethod
    def __iter__(self) -> Generator["ArrayMeta", None, None]:
        """All adapters should provide iterator interface."""
        pass


class BaseArrayAdapter(IArrayAdapter, ABC):
    """Interface for Arrays' adapters."""

    @abstractmethod
    def delete_all_by_vid(self, vid: str, collection: "Collection") -> None:
        """Delete all Arrays in a VArray.

        :param vid: VArray id
        :param collection: Collection instance
        """
        pass


class BaseVArrayAdapter(IArrayAdapter, ABC):
    """Interface for VArrays' adapters."""



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Abstract interfaces for schemas."""

import datetime

from typing import List, Optional, Tuple, Type, Union

import numpy as np

from attr import dataclass

from deker.errors import DekerInvalidSchemaError, DekerValidationError
from deker.tools.schema import get_default_fill_value
from deker.types.private.enums import DTypeEnum
from deker.types.private.typings import Numeric


@dataclass(repr=True)
class BaseSchema:
    """Base schema interface."""

    name: str

    def __attrs_post_init__(self) -> None:
        """Validate after init."""
        if (
            isinstance(self.name, bool)
            or not isinstance(self.name, str)
            or not self.name
            or self.name.isspace()
        ):
            raise DekerValidationError('Dimension schema "name" shall be a non-empty string')

    @property
    def as_dict(self) -> dict:
        """Serialize schema as dict."""
        return {"name": self.name}

    def __str__(self) -> str:
        return self.__repr__()


@dataclass(repr=True)
class BaseAttributeSchema(BaseSchema):
    """Base schema of attribute."""

    dtype: Type[Union[int, float, complex, str, tuple, datetime.datetime]]
    primary: bool


@dataclass(repr=True)
class BaseDimensionSchema(BaseSchema):
    """Dimension base schema interface."""

    size: int

    def __attrs_post_init__(self) -> None:
        """Validate after init."""
        super().__attrs_post_init__()
        if isinstance(self.size, bool) or not isinstance(self.size, int) or not self.size:
            raise DekerValidationError('Dimension schema "size" shall be a positive integer')

    @property
    def as_dict(self) -> dict:
        """Serialize DimensionSchema as dict."""
        d = super().as_dict
        d["size"] = self.size
        return d


@dataclass(repr=True, kw_only=True)
class BaseArraysSchema:
    """Base class for schema."""

    dtype: Type[Numeric]
    fill_value: Union[Numeric, type(np.nan), None]  # type: ignore[valid-type]
    dimensions: Union[List[BaseDimensionSchema], Tuple[BaseDimensionSchema, ...]]
    attributes: Union[List[BaseAttributeSchema], Tuple[BaseAttributeSchema, ...]]

    @property
    def primary_attributes(self) -> Optional[Tuple[BaseAttributeSchema, ...]]:
        """Return only primary attributes."""
        attrs = tuple(attr for attr in self.attributes if attr.primary)
        return attrs if attrs else None

    @property
    def custom_attributes(self) -> Optional[Tuple[BaseAttributeSchema, ...]]:
        """Return only custom attributes."""
        attrs = tuple(attr for attr in self.attributes if not attr.primary)
        return attrs if attrs else None

    def __attrs_post_init__(self) -> None:
        """Validate after init."""
        dimensions_type_error = (
            "Dimensions shall be a list or tuple of DimensionSchemas or TimeDimensionSchemas"
        )
        if not isinstance(self.dimensions, (list, tuple)):
            raise DekerValidationError(dimensions_type_error)
        if not self.dimensions or any(
            not isinstance(d, BaseDimensionSchema) for d in self.dimensions
        ):
            raise DekerValidationError(dimensions_type_error)
        if isinstance(self.dimensions, list):
            self.dimensions = tuple(self.dimensions)

        if len({d.name for d in self.dimensions}) < len(self.dimensions):
            raise DekerValidationError("Dimensions shall have unique names")

        try:
            if self.dtype == int:
                self.dtype = np.int64
            elif self.dtype == float:
                self.dtype = np.float64
            elif self.dtype == complex:
                self.dtype = np.complex128
            DTypeEnum(self.dtype).value  # noqa[B018]
        except ValueError:
            raise DekerValidationError(f"Invalid dtype value {self.dtype}")

        self.fill_value = (
            get_default_fill_value(self.dtype) if self.fill_value is None else self.fill_value
        )

        try:
            fill_value = self.dtype(self.fill_value)  # type: ignore[arg-type]
            np.array(fill_value, self.dtype)
            self.fill_value = fill_value
        except ValueError as e:
            raise DekerValidationError(
                f"Invalid `fill_value` value for dtype {type(self.dtype)}: {e}"
            )

        self.__shape: Tuple[int, ...] = tuple(dim.size for dim in self.dimensions)
        self.__named_shape: Tuple[Tuple[str, int], ...] = tuple(
            (dim.name, dim.size) for dim in self.dimensions
        )

    @property
    def shape(self) -> Tuple[int, ...]:
        """Get array shape."""
        return self.__shape

    @property
    def named_shape(self) -> Tuple[Tuple[str, int], ...]:
        """Get array shape mapping."""
        return self.__named_shape

    @property
    def as_dict(self) -> dict:
        """Serialize as dict."""
        try:
            dtype = DTypeEnum.get_name(DTypeEnum(self.dtype))
            fill_value = None if np.isnan(self.fill_value) else str(self.fill_value)  # type: ignore[arg-type]

            return {
                "dimensions": tuple(dim.as_dict for dim in self.dimensions),
                "dtype": dtype,
                "attributes": tuple(attr.as_dict for attr in self.attributes),
                "fill_value": fill_value,
            }
        except (KeyError, ValueError) as e:
            raise DekerInvalidSchemaError(
                f'Schema "{self.__class__.__name__}" is invalid/corrupted: {e}'
            )



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime

from pathlib import Path
from typing import TYPE_CHECKING, Generator, Optional, Type, Union

from deker.ABC.base_adapters import BaseCollectionAdapter, BaseStorageAdapter
from deker.ABC.base_collection import BaseCollectionOptions
from deker.arrays import Array, VArray
from deker.errors import (
    DekerInstanceNotExistsError,
    DekerInvalidManagerCallError,
    DekerValidationError,
)
from deker.log import SelfLoggerMixin
from deker.managers import ArrayManager, DataManager, FilteredManager, VArrayManager
from deker.schemas import (
    ArraySchema,
    AttributeSchema,
    DimensionSchema,
    SchemaTypeEnum,
    TimeDimensionSchema,
    VArraySchema,
)
from deker.tools import check_memory, not_deleted
from deker.types import Serializer


if TYPE_CHECKING:
    from deker.ABC.base_factory import BaseAdaptersFactory


class Collection(SelfLoggerMixin, Serializer):
    """Collection of ``Arrays`` or ``VArrays``.

    ``Collection`` is a high-level object for managing contents of a set of ``Arrays`` or ``VArrays``
    united into one group under collection name and certain schema.

    Properties
    -----------
    - ``name``: returns ``Collection`` name
    - ``array_schema``: returns schema of embedded ``Arrays``
    - ``varray_schema``: returns schema of ``VArrays`` if applicable, else None
    - ``path``: returns storage path to the ``Collection``
    - ``options``: returns chunking and compression options
    - ``as_dict``: serializes main information about ``Collection`` into dictionary, prepared for JSON

    API methods
    -----------
    - ``create``: according to main schema creates new ``Array`` or ``VArray`` of storage and returns its object
    - ``clear``: according to main schema removes all ``VArrays`` and/or ``Arrays`` from the storage
    - ``delete``: removes ``Collection`` and all its ``VArrays`` and/or ``Arrays`` from the storage
    - ``filter``: filters ``Arrays`` or ``VArrays`` according to main schema and provided conditions
    - ``__iter__``: according to the collection's main schema iterates over all ``Arrays`` or ``VArrays`` in
      ``Collection``, yields their objects
    - ``__str__``: ordinary behaviour
    - ``__repr__``: ordinary behaviour
    """

    __slots__ = (
        "__name",
        "__adapter",
        "__factory",
        "__array_schema",
        "__varray_schema",
        "__manager",
        "__path",
    )

    @staticmethod
    def __create_array_schema_from_varray_schema(schema: VArraySchema) -> ArraySchema:
        """Create ArraySchema from VArraySchema.

        :param schema: VArraySchema instance
        """
        attributes = [
            AttributeSchema(name="vid", dtype=str, primary=True),  # type: ignore[call-arg]
            AttributeSchema(name="v_position", dtype=tuple, primary=True),  # type: ignore[call-arg]
        ]

        dimensions = []
        for n, dim in enumerate(schema.dimensions):
            d = {"name": dim.name, "size": dim.size // schema.vgrid[n]}
            if isinstance(dim, TimeDimensionSchema):
                obj = TimeDimensionSchema
                d["step"] = dim.step
                d["start_value"] = dim.start_value

                if isinstance(dim.start_value, str) and dim.start_value.startswith("$"):
                    d["start_value"] = f"$parent.{dim.start_value[1:]}"
                    attributes.append(
                        AttributeSchema(  # type: ignore[call-arg]
                            name=d["start_value"][1:],
                            dtype=datetime.datetime,
                            primary=False,
                        )
                    )
            else:
                obj = DimensionSchema
                d["labels"] = None
                d["scale"] = None

            dimensions.append(obj(**d))  # type: ignore[arg-type]
        return ArraySchema(  # type: ignore[call-arg]
            dtype=schema.dtype,
            fill_value=schema.fill_value,
            dimensions=dimensions,  # type: ignore[arg-type]
            attributes=attributes,
        )

    def _validate(
        self,
        name: str,
        schema: Union[ArraySchema, VArraySchema],
    ) -> None:
        """Validate name and schema.

        :param name: Collection name
        :param schema: Collection main schema
        """
        if not name or not isinstance(name, str) or name.isspace():
            raise DekerValidationError("Collection name shall be a non-empty string")

        if schema is None or not isinstance(schema, (ArraySchema, VArraySchema)):
            raise DekerValidationError(
                f"Invalid schema type {type(schema)}; ArraySchema or VArraySchema expected"
            )

    def __init__(
        self,
        name: str,
        schema: Union[ArraySchema, VArraySchema],
        adapter: "BaseCollectionAdapter",
        factory: "BaseAdaptersFactory",
        storage_adapter: Type["BaseStorageAdapter"],
        collection_options: Optional[BaseCollectionOptions] = None,
    ) -> None:
        """Collection constructor.

        :param name: ``Collection`` unique name
        :param schema: ``ArraySchema`` or ``VArraySchema`` instance
        :param adapter: ``CollectionAdapter`` instance
        :param factory: ``AdaptersFactory`` instance
        :param storage_adapter: Storage adapter type
        :param collection_options: ``CollectionOptions`` instance or None
        """
        self._validate(name, schema)
        self.__name: str = name
        self.__adapter: BaseCollectionAdapter = adapter
        self.__factory = factory
        self.__path = self.__adapter.collections_resource / name  # type: ignore[attr-defined]
        # Init collection chunking and compression options
        self.__collection_options = collection_options
        self.__storage_adapter = storage_adapter
        array_adapter = self.__factory.get_array_adapter(
            self.__path, self.__storage_adapter, self.__collection_options
        )

        if isinstance(schema, VArraySchema):
            self.__array_schema: ArraySchema = self.__create_array_schema_from_varray_schema(schema)
            self.__varray_schema: VArraySchema = schema
            varray_adapter = self.__factory.get_varray_adapter(
                self.__path, storage_adapter=self.__storage_adapter
            )
            self.__varrays = VArrayManager(self, array_adapter, varray_adapter)
        else:
            self.__array_schema: ArraySchema = schema
            self.__varray_schema: None = None
            varray_adapter = None
            self.__varrays = None

        self.__arrays = ArrayManager(
            self,
            array_adapter,
        )

        self.__manager: DataManager = DataManager(self, array_adapter, varray_adapter)
        self.logger.debug(f"Collection {self.__name} initialized")

    def __del__(self) -> None:
        """Delete object."""
        del self.__adapter
        del self.__manager
        del self.__arrays
        del self.__varrays

    @property
    def _storage_adapter(self) -> Type[BaseStorageAdapter]:
        """Return type of storage adapter."""
        return self.__storage_adapter

    @property
    def name(self) -> str:
        """Get collection name."""
        return self.__name

    @property
    def array_schema(self) -> ArraySchema:
        """Get collection ``ArraySchema``."""
        return self.__array_schema

    @property
    def varray_schema(self) -> Optional[VArraySchema]:
        """Get collection ``VArraySchema``."""
        return self.__varray_schema

    @property
    def path(self) -> Path:
        """Get ``Collection`` fs-path."""
        return self.__path

    @property
    def options(self) -> Optional[BaseCollectionOptions]:
        """Get collection options: chunking, compression, etc."""
        return self.__collection_options

    @property
    def as_dict(self) -> dict:
        """Serialize ``Collection`` to dictionary."""
        dic: dict = {
            "name": self.name,
            "type": SchemaTypeEnum.varray.name
            if self.__varray_schema
            else SchemaTypeEnum.array.name,
            "schema": self.varray_schema.as_dict
            if self.__varray_schema
            else self.array_schema.as_dict,
            "options": self.options.as_dict if self.options else self.options,
            "storage_adapter": self.__storage_adapter.__name__,
            "metadata_version": self.__adapter.metadata_version,
        }

        return dic

    @property
    def varrays(self) -> Optional[VArrayManager]:
        """Return manager for ``VArrays``."""
        if not self.__varrays:
            raise DekerInvalidManagerCallError(
                "``.varrays`` in `Arrays` collection is unavailable; use ``.arrays`` instead"
            )
        return self.__varrays

    @property
    def arrays(self) -> ArrayManager:
        """Return manager for ``Arrays``."""
        return self.__arrays

    @not_deleted
    def create(
        self, primary_attributes: Optional[dict] = None, custom_attributes: Optional[dict] = None
    ) -> Union[Array, VArray]:
        """Create ``Array`` or ``VArray`` according to collection main schema.

        If ``VArraySchema`` is passed to ``Collection``, all data management will go through ``VArrays``
        as this method will create just ``VArrays`` (``Arrays`` will be created automatically by ``VArray``).
        Otherwise, only ``Arrays`` will be created.

        :param primary_attributes: ``Array`` or ``VArray`` primary attribute
        :param custom_attributes: ``VArray`` or ``VArray`` custom attributes
        """
        schema = self.array_schema
        shape = schema.arrays_shape if hasattr(schema, "arrays_shape") else schema.shape
        check_memory(shape, schema.dtype, self.__adapter.ctx.config.memory_limit)
        array = self.__manager.create(primary_attributes, custom_attributes)
        self.logger.debug(
            f"{array.__class__.__name__} id={array.id} {primary_attributes=}, {custom_attributes=} created"
        )
        return array

    def delete(self) -> None:
        """Remove ``Collection`` and all its contents from the database."""
        self.logger.debug(f"Removing collection {self.__name} from database")
        self.__adapter.delete(self)
        self.logger.debug(f"Collection {self.__name} deleted OK")

    @not_deleted
    def clear(self) -> None:
        """Clear all ``Arrays`` and/or ``VArrays``  inside ``Collection``."""
        self.logger.debug(f"Clearing data from collection {self.__name}")
        self.__adapter.clear(self)
        self.logger.debug(f"Collection {self.__name} cleared OK")

    def filter(self, filters: dict) -> FilteredManager:
        """Filter ``Arrays`` or ``VArrays`` by provided conditions.

        :param filters: query conditions for filtering

        .. note::
           Conditions for filtering are either Array or VArray ``id`` value::

             {"id": "some_array_UUID_string"}

           or full scope of primary attributes' values::

             {"primary_attr1_name": its value, "primary_attr2_name": its value, ...}
        """
        return self.__manager.filter(filters)

    def _is_deleted(self) -> bool:
        """Check if collection was deleted."""
        return self.__adapter.is_deleted(self)  # type: ignore[attr-defined]

    def __next__(self) -> None:
        pass

    def __iter__(self) -> Generator[Union[Array, VArray], None, None]:
        """Yield ``VArrays`` or ``VArrays`` from the ``Collection``."""
        if self._is_deleted():
            raise DekerInstanceNotExistsError(
                f"{self} doesn't exist, create new or get an instance again to be able to iterate"
            )
        array_adapter = self.__factory.get_array_adapter(
            self.path, self.__storage_adapter, self.options
        )
        varray_adapter = (
            self.__factory.get_varray_adapter(self.path, storage_adapter=self.__storage_adapter)
            if self.varray_schema
            else None
        )

        adapter = varray_adapter if varray_adapter else array_adapter

        self.logger.debug(f"Iterating over collection {self.name}")
        for meta in adapter:
            if varray_adapter:
                array: VArray = VArray._create_from_meta(
                    self, meta, array_adapter, varray_adapter=varray_adapter
                )
            else:
                array: Array = Array._create_from_meta(self, meta, array_adapter)
            yield array

    def __repr__(self) -> str:
        s = f"{self.__class__.__name__}(name={self.name}, array_schema={self.array_schema})"
        if self.varray_schema:
            s = s[:-1] + f", varray_schema={self.varray_schema})"
        return s

    def __str__(self) -> str:
        return f"{self.name}"



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Abstract wrappers for array dimension and dimension index labels."""

from abc import ABC
from typing import Union

from deker.errors import DekerValidationError
from deker.types.private.classes import Serializer


class BaseDimension(Serializer, ABC):
    """Dimension abstract object providing all its inheritors with common actions and methods."""

    __slots__ = ("__name", "__size", "__step")

    def _validate(self, name: str, size: int, **kwargs: dict) -> None:  # noqa[ARG002]
        if not isinstance(name, str):
            raise DekerValidationError("Name shall be str")
        if not name or name.isspace():
            raise DekerValidationError("Name cannot be empty")
        if size is None or isinstance(size, bool) or not isinstance(size, int) or size <= 0:
            raise DekerValidationError("Size shall be a positive int")

    def __init__(self, name: str, size: int, **kwargs: dict) -> None:
        super().__init__()
        self._validate(name, size, **kwargs)  # pragma: no cover
        self.__name = name  # pragma: no cover
        self.__size = size  # pragma: no cover
        self.__step = 1  # pragma: no cover

    @property
    def name(self) -> str:
        """Dimension name."""
        return self.__name  # pragma: no cover

    @property
    def size(self) -> int:
        """Dimension size."""
        return self.__size  # pragma: no cover

    @property
    def step(self) -> Union[int, float]:
        """Dimension values step."""
        return self.__step  # pragma: no cover



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from datetime import datetime
from typing import Optional


def convert_datetime_attrs_to_iso(attrs: Optional[dict]) -> Optional[dict]:
    """Convert datetime attributes to iso-format if possible.

    :param attrs: attributes dict
    """
    if attrs is not None and not isinstance(attrs, dict):
        raise TypeError(f"Invalid attribute value type: {type(attrs)}; expected dict")

    if attrs:
        return {
            key: (value.isoformat() if isinstance(value, datetime) else value)
            for key, value in attrs.items()
        }
    return attrs


def convert_iso_attrs_to_datetime(attrs: Optional[dict]) -> Optional[dict]:
    """Convert iso-format attributes to datetime if possible.

    :param attrs: attributes dict
    """
    if attrs is not None and not isinstance(attrs, dict):
        raise TypeError(f"Invalid attribute value type: {type(attrs)}; expected dict")

    if attrs:
        new_attrs = {}
        for key, value in attrs.items():
            if isinstance(value, str):
                try:
                    new_attrs[key] = datetime.fromisoformat(value)
                except (TypeError, ValueError):
                    new_attrs[key] = value
            else:
                new_attrs[key] = value
        return new_attrs
    return attrs



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from collections import OrderedDict
from pathlib import Path
from typing import Dict, List, Union
from urllib.parse import ParseResult, parse_qs, quote, urlparse

from deker_tools.path import is_path_valid

from deker.errors import DekerValidationError


class Uri(ParseResult):
    """Deker client uri wrapper.

    Overrides parsed result to have query as a dictionary.
    Validation ref: https://t.ly/ekbU
    """

    __annotations__ = OrderedDict(
        scheme={"separator": None, "divider": None},
        netloc={"separator": "://", "divider": None},
        path={"separator": "/", "divider": None},
        params={"separator": ";", "divider": None},
        query={"separator": "?", "divider": "&"},
        fragment={"separator": "#", "divider": None},
    )
    query: Dict[str, List[str]]

    @property
    def raw_url(self) -> str:
        """Get url from raw uri without query string, arguments and fragments."""
        url = self.scheme + "://"
        if self.netloc:
            url += self.netloc
        url += quote(str(self.path), safe=":/")
        return url

    @classmethod
    def __parse(cls, uri: str) -> Uri:
        """Parse uri from string.

        :param uri: : scheme://username:password@host:port/path;parameters?query
        """
        result = urlparse(uri)

        if ";" in result.path:
            path, params = result.path.split(";")
        else:
            path, params = result.path, result.params

        query = parse_qs(result.query)

        return Uri(  # type: ignore
            result.scheme,
            result.netloc,
            path,  # type: ignore[arg-type]
            params,
            query,  # type: ignore[arg-type]
            result.fragment,
        )

    @classmethod
    def validate(cls, uri: str) -> None:
        """Validate uri from string.

        :param uri: : scheme://username:password@host:port/path;parameters?query
        """
        if not isinstance(uri, str):
            raise DekerValidationError(f"Invalid uri type: {uri} - {type(uri)}")
        if not uri or uri.isspace():
            raise DekerValidationError("Empty uri passed")

        parsed_uri = cls.__parse(uri)
        if parsed_uri.scheme == "file":
            pathname = Path(parsed_uri.path)
            try:
                is_path_valid(pathname)
            except Exception as e:
                raise DekerValidationError(f"Invalid uri: {e}; {uri}")

    @classmethod
    def create(cls, uri: str) -> Uri:
        """Create parsed and validated uri from string.

        :param uri: : scheme://username:password@host:port/path;parameters?query
        """
        cls.validate(uri)
        return cls.__parse(uri)

    def __truediv__(self, other: Union[str, Path]) -> Uri:
        """Join uri to Path or string.

        :param other: Path or string to join
        """
        sep = "/"
        other = str(other)
        path = sep.join((self.path, other.strip()))
        res = Uri(  # type: ignore
            self.scheme,
            self.netloc,
            path,  # type: ignore[arg-type]
            self.params,
            self.query,  # type: ignore[arg-type]
            self.fragment,
        )
        return res

    def __itruediv__(self, other: str) -> Uri:
        return self.__truediv__(other)



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List, NamedTuple, Optional, Tuple, Union

from deker.errors import DekerValidationError
from deker.log import SelfLoggerMixin


Labels = Union[Tuple[Union[str, int, float], ...], List[Union[str, int, float]]]


class Scale(NamedTuple):
    """NamedTuple with a description of a dimension regular scale.

    For example we describe dimensions for the Earth's latitude and longitude grades::

        dims = [
                DimensionSchema(name="y", size=721),
                DimensionSchema(name="x", size=1440),
            ]

    Such description may exist, but it's not quite sufficient. We can provide more information for the grid::

        dims = [
                    DimensionSchema(
                        name="y",
                        size=721,
                        scale=Scale(start_value=90, step=-0.25, name="lat")
                    ),
                    DimensionSchema(
                        name="x",
                        size=1440,
                        scale=Scale(start_value=-180, step=0.25, name="lon")
                    ),
                ]

    This extra information permits us provide fancy indexing by lat/lon coordinates in degrees::

        EarthGridArray[1, 1] == EarthGridArray[89.75, -179.75]

    :param start_value: scale real start point
    :param step: scale values step (may be positive or negative)
    :param name: optional scale name (latitude, longitude or whatever)
    """

    start_value: float
    step: float
    name: Optional[str] = None


class StorageSize(NamedTuple):
    """Namedtuple for storage size with bytes and human representation."""

    bytes: int
    human: str


class IndexLabels(tuple, SelfLoggerMixin):
    """``IndexLabels`` is an overloaded ``tuple`` providing a mapping of label-values to their ``Dimension`` indexes.

    It provides direct and reversed ordered mappings of dimension's values names (labels)
    to their position in the axis flat array (indexes).

    Labels shall be a sequence of unique values within the full scope of the passed ``Dimension``.
    Valid Labels are list or tuple of unique strings or floats::

      ["name1", "name2", ..., "nameN"] | (0.2, 0.1, 0.4, 0.3, ..., -256.78963)

    :param labels: list or tuple of unique strings or floats
    """

    @classmethod
    def __validate(cls, labels: Optional[Labels]) -> Tuple[Union[str, int, float], ...]:
        error = DekerValidationError(
            "Labels shall be a list or tuple of unique values of the same type (strings, floats), "
            "not None or empty list or tuple"
        )
        if (
            not labels
            or not isinstance(labels, (list, tuple))
            or all(
                (
                    any(not isinstance(val, str) for val in labels),
                    any(not isinstance(val, float) for val in labels),
                )
            )
        ):
            raise error
        return labels if isinstance(labels, tuple) else tuple(labels)

    def __new__(cls, labels: Optional[Labels]) -> IndexLabels:
        """Override python ``tuple.__new__`` method.

        Here we add values validation before an instance is created.

        :param labels: labels
        """
        return super(IndexLabels, cls).__new__(cls, cls.__validate(labels))  # type: ignore[arg-type]

    def __init__(self, labels: Optional[Labels]) -> None:
        self.logger.debug("labels initialized")

    @property
    def first(self) -> Union[str, int, float]:
        """Get first label."""
        return self[0]

    @property
    def last(self) -> Union[str, int, float]:
        """Get last label."""
        return self[-1]

    def name_to_index(self, name: Union[str, int, float]) -> Optional[int]:
        """Get label index by its name.

        :param name: label name
        """
        try:
            return self.index(name)
        except ValueError as e:
            mes = str(e) + f": no name {name}"
            self.logger.debug(mes)
            return None

    def index_to_name(self, idx: int) -> Optional[Union[str, int, float]]:
        """Get label name by its index.

        :param idx: label index
        """
        try:
            return self[idx]
        except IndexError as e:
            mes = str(e) + f": no index {idx}"
            self.logger.debug(mes)
            return None

    def __str__(self) -> str:
        return str(tuple(self))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__str__()})"



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# nopycln: file
from .array import (
    calculate_total_cells_in_array,
    check_memory,
    convert_human_memory_to_bytes,
    get_id,
)
from .decorators import not_deleted
from .path import get_array_lock_path, get_main_path, get_paths, get_symlink_path
from .schema import (
    create_attributes_schema,
    create_dimensions,
    create_dimensions_schema,
    get_default_fill_value,
)
from .time import convert_datetime_attrs_to_iso, convert_iso_attrs_to_datetime



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple, Tuple, Union

from typing_extensions import NotRequired, TypedDict


if TYPE_CHECKING:
    from deker.types import Data

__all__ = (
    "Serializer",
    "ArrayMeta",
    "ArrayLockMeta",
    "ArrayPosition",
    "ArrayOffset",
    "ArraysCoordinatesWithOffset",
    "CollectionLockMeta",
    "Paths",
)


class Serializer(ABC):
    """Base abstract object with a possibility to serialize itself into different objects."""

    @property
    @abstractmethod
    def as_dict(self) -> dict:
        """Serialize self into a dictionary."""
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass


class ArrayMeta(TypedDict):
    """Typing for array meta."""

    id: str
    primary_attributes: dict
    custom_attributes: dict


class Paths(NamedTuple):
    """Namedtuple for array paths."""

    main: Path
    symlink: Path

    def create(self) -> None:
        """Create paths in the file system."""
        for attr in self._fields:
            path: Path = getattr(self, attr)
            # TODO: make universal string check for is_file. "path" is str, not Path
            if str(path).endswith("hdf5") or str(path).endswith("json"):
                path = path.parent
            path.mkdir(parents=True, exist_ok=True)


class ArrayOffset(TypedDict):
    """Offset of subset in array."""

    start: int
    end: NotRequired[int]


class ArraysCoordinatesWithOffset(NamedTuple):
    """Coordinate of Array in dimension, with offset."""

    coordinate: int
    offset: ArrayOffset


class ArrayPosition(NamedTuple):
    """Tuple with array coordinates, bounds and slice that will be passed to adapter."""

    vposition: Tuple[int, ...]
    bounds: Tuple[Union[slice, int], ...]
    data_slice: Tuple[slice, ...]


class ArrayPositionedData(NamedTuple):
    """Tuple with array coordinates, bounds and data that will be passed to adapter."""

    vposition: Tuple[int, ...]
    bounds: Tuple[Union[slice, int], ...]
    data: "Data"


class CollectionLockMeta(TypedDict):
    """Typed dict describing collection lockfile parameters. Used for typing."""

    Lockfile: str
    Collection: str
    Type: str
    Creation: str


class ArrayLockMeta(TypedDict):
    """Typed dict describing array lockfile parameters. Used for typing."""

    Lockfile: str
    Collection: str
    Array: str
    ID: str
    PID: str
    TID: str
    Type: str
    Creation: str



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Abstract interfaces for integrity checker."""

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Type


if TYPE_CHECKING:
    from deker.client import Client


class BaseChecker(ABC):
    """Checker abstract object. Implements chain of responsibility pattern for integrity check."""

    def __init__(
        self,
        stop_on_error: bool,
        paths: dict,
        errors: dict,
        level: int,
        client: "Client",
        root_path: Path,
    ):
        super().__init__()
        self.stop_on_error = stop_on_error
        self.paths = paths
        self.errors = errors
        self.level = level
        self.client = client
        self.root_path = root_path
        self.ctx = client._Client__ctx  # type: ignore
        self.next_checker = None

    def _parse_errors(self) -> str:
        """Parse self.errors and return string."""
        res = ""
        if not self.errors:
            return res
        for key in self.errors:
            res += key + "\n\t- " + "\n\t- ".join(self.errors[key]) + "\n"
        return res

    def add_checker(self, checker: Type["BaseChecker"]) -> None:
        """Add next checker, uses self to init.

        :param checker: next_checker type
        """
        current = self
        while current.next_checker:
            current = current.next_checker
        current.next_checker = checker(
            self.stop_on_error, self.paths, self.errors, self.level, self.client, self.root_path
        )

    @abstractmethod
    def check(self, *args: Any, **kwargs: Any) -> None:
        """Check integrity. Shall be implemented in every Checker.

        :param args: args
        :param kwargs: kwargs
        """
        pass



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Abstract interfaces for factories."""

from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from typing import TYPE_CHECKING, Any, Tuple

from deker.log import SelfLoggerMixin


if TYPE_CHECKING:
    from deker.ABC.base_adapters import BaseArrayAdapter, BaseCollectionAdapter, BaseVArrayAdapter
    from deker.ctx import CTX
    from deker.uri import Uri


class BaseAdaptersFactory(SelfLoggerMixin, ABC):
    """Base interface of Collections, Arrays and VArrays adapters' factory."""

    uri_schemes: Tuple[str, ...]

    def __init__(self, ctx: "CTX", uri: "Uri") -> None:
        self.ctx = ctx
        self.uri = uri
        if self.ctx.executor is None:
            self.executor: ThreadPoolExecutor = ThreadPoolExecutor(self.ctx.config.workers)
            self.ctx.executor = self.executor
            self._own_executor = True
        else:
            self.executor: ThreadPoolExecutor = self.ctx.executor
            self._own_executor: bool = False
        self.logger.debug(f"{self.__class__.__name__} instantiated")

    def __del__(self) -> None:
        self.close()

    @abstractmethod
    def close(self) -> None:
        """Close own executor."""
        try:
            if self.executor and self._own_executor:
                self.executor.shutdown(wait=True, cancel_futures=False)
        except AttributeError:
            pass

    @abstractmethod
    def get_array_adapter(self, *args: Any, **kwargs: Any) -> "BaseArrayAdapter":
        """Create ArrayAdapter instance.

        :param args: any arguments
        :param kwargs: any keyword arguments
        """
        pass

    @abstractmethod
    def get_varray_adapter(self, *args: Any, **kwargs: Any) -> "BaseVArrayAdapter":  # type: ignore[return-value]
        """Create VArrayAdapter instance.

        :param args: any arguments
        :param kwargs: any keyword arguments
        """
        pass

    @abstractmethod
    def get_collection_adapter(self, *args: Any, **kwargs: Any) -> "BaseCollectionAdapter":
        """Create collection adapter.

        :param args: any arguments
        :param kwargs: any keyword arguments
        """
        pass
